#include "nvfs.h"

static struct nvfs_dir_entry *nvfs_prepare_page(void *map, struct dentry *dentry)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(d_inode(dentry->d_parent));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	void *end = map + (1UL << nvs->log2_page_size);
	size_t needed_len = nvfs_needed_size(dentry->d_name.len);
	do {
		struct nvfs_dir_entry *nde = map;
		size_t de_size = nvfs_dir_entry_size(nde);
		size_t nns;

		if (le32_to_cpu(nde->name_len_and_type) & TYPE_MASK)
			goto used;

		if (nvfs_dir_entry_size(nde) < needed_len) {
			size_t next_size;
			struct nvfs_dir_entry *next_nde = map + de_size;
			if ((void *)next_nde >= end)
				goto used;
			if (le32_to_cpu(next_nde->name_len_and_type) & TYPE_MASK)
				goto used;
			next_size = de_size + nvfs_dir_entry_size(next_nde);
			nv_store(nde->name_len_and_type, cpu_to_le32(nvfs_make_size(next_size) << NAME_LEN_OFFSET));
			wmb();
			continue;
		}

		nns = nvfs_needed_size(le32_to_cpu(nde->name_len_and_type));
		if (nns > needed_len) {
			size_t remaining;
			struct nvfs_dir_entry *next_nde;
			remaining = nns - needed_len;
			if (unlikely(remaining < 16))
				goto used;
			next_nde = map + needed_len;
			nv_store(next_nde->name_len_and_type, cpu_to_le32(nvfs_make_size(remaining) << NAME_LEN_OFFSET));
		}

		if (dentry->d_name.len < 4) {
			__le32 key = cpu_to_le32(0);
			memcpy(&key, dentry->d_name.name, dentry->d_name.len);
			nv_store(*(__le32 *)nde->name, key);
		} else {
			memcpy_flushcache(nde->name, dentry->d_name.name, dentry->d_name.len);
		}
		nv_store(nde->name_len_and_type, cpu_to_le32((dentry->d_name.len << NAME_LEN_OFFSET)));
		return nde;

used:
		map += de_size;
	} while (map < end);
	if (unlikely(map > end)) {
		nvfs_error(nvs, "corrupted directory page");
		return ERR_PTR(-EUCLEAN);
	}
	return (void *)1;
}

static void init_dir_page(struct nvfs_superblock *nvs, void *page, void *aux)
{
	struct nvfs_dir_entry *nde = page;
	nv_store(nde->name_len_and_type, cpu_to_le32(nvfs_make_size(1UL << nvs->log2_page_size) << NAME_LEN_OFFSET));
}

struct nvfs_dir_entry *nvfs_prepare_add(struct dentry *dentry)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(d_inode(dentry->d_parent));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	uint64_t page, limit;
	struct nvfs_dir_entry *nde;

	limit = nmi->vfs_inode.i_size >> nvs->log2_page_size;
	for (page = 0; page < limit; page++) {
		void *map = nvfs_bmap(nmi, page, NULL, NULL, NULL);
		if (map) {
			nde = nvfs_prepare_page(map, dentry);
			if (IS_ERR(nde))
				return nde;
			if (nde == (void *)1)
				continue;
			return nde;
		}
	}
	for (page = 0; page < limit + 1; page++) {
		bool new;
		void *map = nvfs_bmap(nmi, page, init_dir_page, NULL, &new);
		if (new) {
			if (likely(page == limit)) {
				nvfs_set_inode_size(nmi, nmi->vfs_inode.i_size + (1UL << nvs->log2_page_size));
			}
			nde = nvfs_prepare_page(map, dentry);
			if (IS_ERR(nde))
				return nde;
			if (nde == (void *)1)
				continue;
			return nde;
		}
	}

	return ERR_PTR(-ENOSPC);
}

void nvfs_finish_add(struct nvfs_dir_entry *nde, struct dentry *dentry, struct inode *inode)
{
	nv_store(nde->inode, cpu_to_le64(inode->i_ino));
	wmb();
	nv_store(nde->name_len_and_type, cpu_to_le32((dentry->d_name.len << NAME_LEN_OFFSET) | (fs_umode_to_ftype(inode->i_mode) + 1) << TYPE_OFFSET));
}

int nvfs_add_link(struct dentry *dentry, struct inode *inode)
{
	struct nvfs_dir_entry *nde;

	nde = nvfs_prepare_add(dentry);

	if (unlikely(IS_ERR(nde)))
		return PTR_ERR(nde);

	nvfs_finish_add(nde, dentry, inode);

	return 0;
}

struct nvfs_dir_entry *nvfs_find_de(struct nvfs_memory_inode *nmi, struct qstr *qname)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	uint64_t page, limit;

	__le32 key = cpu_to_le32(0);
	memcpy(&key, qname->name, min(4U, qname->len));

	limit = nmi->vfs_inode.i_size >> nvs->log2_page_size;
	for (page = 0; page < limit; page++) {
		void *map = nvfs_bmap(nmi, page, NULL, NULL, NULL);
		if (map) {
			void *end = map + (1UL << nvs->log2_page_size);
			while (map < end) {
				struct nvfs_dir_entry *nde = map;

				if (likely(*(__le32 *)nde->name != key))
					goto next;

				if (unlikely(!nvfs_dir_entry_type(nde)))
					goto next;

				if (qname->len != nvfs_dir_entry_namelen(nde))
					goto next;

				if (memcmp(nde->name, qname->name, qname->len))
					goto next;

				return nde;

next:
				map += nvfs_dir_entry_size(nde);
			}
			if (unlikely(map > end)) {
				nvfs_error(nvs, "corrupted directory page");
				return ERR_PTR(-EUCLEAN);
			}
		}
	}

	return NULL;
}

int nvfs_delete_de(struct nvfs_memory_inode *nmi, struct qstr *qname)
{
	struct nvfs_dir_entry *nde;
	uint32_t name_len_and_type;

	nde = nvfs_find_de(nmi, qname);
	if (unlikely(IS_ERR(nde)))
		return PTR_ERR(nde);
	if (!nde)
		return -ENOENT;

	name_len_and_type = le32_to_cpu(nde->name_len_and_type);
	name_len_and_type &= NAME_LEN_MASK;
	nde->name_len_and_type = cpu_to_le32(name_len_and_type);

	return 0;
}

static struct dentry *nvfs_lookup(struct inode *dir, struct dentry *dentry, unsigned int flags)
{
	struct inode *inode;
	struct nvfs_dir_entry *nde;
	nde = nvfs_find_de(i_to_nmi(dir), &dentry->d_name);
	if (!nde)
		return d_splice_alias(NULL, dentry);
	if (IS_ERR(nde))
		return (void *)nde;
	inode = nvfs_iget(dir->i_sb, le64_to_cpu(nde->inode));
	if (unlikely(inode == ERR_PTR(-ESTALE))) {
		nvfs_error(dir->i_sb->s_fs_info, "deleted inode referenced");
		return ERR_PTR(-EUCLEAN);
	}
	return d_splice_alias(inode, dentry);
}

static int nvfs_readdir(struct file *file, struct dir_context *ctx)
{
	struct inode *inode = file_inode(file);
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	uint64_t page, offset, limit;
	uint64_t pos;

	if (ctx->pos == 0) {
		if (unlikely(!dir_emit_dot(file, ctx)))
			goto ret_0;
		ctx->pos++;
	}
	if (ctx->pos == 1) {
		if (unlikely(!dir_emit_dotdot(file, ctx)))
			goto ret_0;
		ctx->pos++;
	}
	pos = ctx->pos - 2;

	limit = inode->i_size >> nvs->log2_page_size;

	page = pos >> nvs->log2_page_size;
	offset = pos & ((1UL << nvs->log2_page_size) - 1);

	for (; page < limit; page++, offset = 0) {
		void *map = nvfs_bmap(nmi, page, NULL, NULL, NULL);
		if (map) {
			void *start = map;
			void *end = map + (1UL << nvs->log2_page_size);
			while (map < end) {
				struct nvfs_dir_entry *nde = map;
				unsigned type;
				size_t namelen;

				if (map - start < offset)
					goto next;

				offset = map - start;

				type = nvfs_dir_entry_type(nde);
				if (unlikely(!type))
					goto next;

				namelen = nvfs_dir_entry_namelen(nde);

				if (!dir_emit(ctx, nde->name, namelen, le64_to_cpu(nde->inode), fs_ftype_to_dtype(type - 1)))
					goto ret_0_pos;

next:
				map += nvfs_dir_entry_size(nde);
			}
			if (unlikely(map > end)) {
				nvfs_error(nvs, "corrupted directory page");
				return -EUCLEAN;
			}
		}
	}

ret_0_pos:
	pos = ((uint64_t)page << nvs->log2_page_size) + offset;
	ctx->pos = pos + 2;
ret_0:
	return 0;
}

int nvfs_is_dir_empty(struct nvfs_memory_inode *nmi)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	uint64_t page, limit;

	limit = nmi->vfs_inode.i_size >> nvs->log2_page_size;

	for (page = 0; page < limit; page++) {
		void *map = nvfs_bmap(nmi, page, NULL, NULL, NULL);
		if (map) {
			void *end = map + (1UL << nvs->log2_page_size);
			while (map < end) {
				struct nvfs_dir_entry *nde = map;
				if (unlikely(!nvfs_dir_entry_type(nde)))
					goto next;
				return 0;
next:
				map += nvfs_dir_entry_size(nde);
			}
			if (unlikely(map > end)) {
				nvfs_error(nvs, "corrupted directory page");
				return -EUCLEAN;
			}
		}
	}

	return 1;
}

const struct inode_operations nvfs_dir_inode_operations = {
	.lookup = nvfs_lookup,
	.create = nvfs_create,
	.mkdir = nvfs_mkdir,
	.symlink = nvfs_symlink,
	.mknod = nvfs_mknod,
	.unlink = nvfs_unlink,
	.rmdir = nvfs_rmdir,
	.rename = nvfs_rename,
	.link = nvfs_link,
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
};

const struct file_operations nvfs_dir_fops = {
	.llseek = generic_file_llseek,
	.read = generic_read_dir,
	.iterate_shared = nvfs_readdir,
};
