#include "nvfs.h"

static inline ssize_t nvfs_read_locked(struct file *file, char __user *buf, size_t len, loff_t *ppos)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(file->f_inode);
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t pos;
	int r;
	size_t i_size;

	if (unlikely(!len))
		return 0;

	pos = *ppos;

	i_size = nmi->vfs_inode.i_size;
	if (pos >= i_size)
		return 0;
	if (pos + len > i_size)
		len = i_size - pos;

	if (unlikely(!access_ok(buf, len)))
		return -EFAULT;

	do {
		void *blk;

		unsigned page_offset = pos & ((1UL << nvs->log2_page_size) - 1);
		unsigned size = min(len, (1UL << nvs->log2_page_size) - page_offset);

		blk = nvfs_bmap(nmi, pos >> nvs->log2_page_size, NULL, NULL, NULL);
		if (unlikely(!blk)) {
			size = min(size, (unsigned)PAGE_SIZE);
			size -= __copy_to_user(buf, empty_zero_page, size);
		} else {
			size -= __copy_to_user(buf, blk + page_offset, size);
		}

		if (unlikely(!size)) {
			r = -EFAULT;
			goto ret_r;
		}

		buf += size;
		len -= size;
		pos += size;
		total += size;
	} while (len);

	r = 0;

ret_r:
	*ppos = pos;

	file_accessed(file);

	return total ? total : r;
}

struct nvfs_aux {
	const char __user *user_buf;
	unsigned page_offset;
	unsigned size;
	unsigned copied;
	bool clear;
};

static void nvfs_write_page(struct nvfs_superblock *nvs, void *page, void *aux_)
{
	struct nvfs_aux *aux = aux_;

	if (aux->clear)
		nvfs_memzero_flushcache(page, aux->page_offset);

	/* !!! FIXME: this doesn't flush trailing bytes */
	aux->copied = aux->size - __copy_from_user_inatomic_nocache(page + aux->page_offset, aux->user_buf, aux->size);

	if (aux->clear)
		nvfs_memzero_flushcache(page + aux->page_offset + aux->size, (1UL << nvs->log2_page_size) - (aux->page_offset + aux->size));
}

static inline ssize_t nvfs_write_locked(struct file *file, const char __user *buf, size_t len, loff_t *ppos)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(file->f_inode);
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	ssize_t total = 0;
	loff_t pos;
	int r;
	bool some_new = false;

	if (unlikely(!len))
		return 0;

	if (unlikely(!access_ok(buf, len)))
		return -EFAULT;

	r = file_modified(file);
	if (unlikely(r < 0))
		return r;

	if (!(file->f_flags & O_APPEND))
		pos = *ppos;
	else
		pos = nmi->vfs_inode.i_size;

	do {
		struct nvfs_aux aux;
		void *blk;
		bool new;

		aux.page_offset = pos & ((1UL << nvs->log2_page_size) - 1);
		aux.size = min(len, (1UL << nvs->log2_page_size) - aux.page_offset);
		aux.user_buf = buf;
		aux.clear = true;

		blk = nvfs_bmap(nmi, pos >> nvs->log2_page_size, nvfs_write_page, &aux, &new);
		if (unlikely(!blk)) {
			r = -ENOSPC;
			goto ret_r;
		}
		some_new |= new;
		if (!new) {
			aux.clear = false;
			nvfs_write_page(nvs, blk, &aux);
		}

		if (unlikely(!aux.copied)) {
			r = -EFAULT;
			goto ret_r;
		}

		buf += aux.copied;
		len -= aux.copied;
		pos += aux.copied;
		total += aux.copied;
	} while (len);

	r = 0;

ret_r:
	if (some_new && unlikely(nmi->was_mmapped)) {
		loff_t start_pos;
		if (!(file->f_flags & O_APPEND))
			start_pos = *ppos;
		else
			start_pos = nmi->vfs_inode.i_size;
		if (pos > start_pos) {
			uint64_t start_fs_page, end_fs_page, start_page, end_page;

			start_fs_page = start_pos >> nvs->log2_page_size;
			end_fs_page = (pos + (1UL << nvs->log2_page_size) - 1) >> nvs->log2_page_size;

			start_page = start_fs_page << (nvs->log2_page_size - PAGE_SHIFT);
			end_page = end_fs_page << (nvs->log2_page_size - PAGE_SHIFT);

			invalidate_inode_pages2_range(nmi->vfs_inode.i_mapping, start_page, end_page - 1);
		}
	}
	if (pos > nmi->vfs_inode.i_size) {
		wmb();
		nvfs_set_inode_size(nmi, pos);
	}

	if (!(file->f_flags & O_APPEND))
		*ppos = pos;

	return total ? total : r;
}

static ssize_t nvfs_read(struct file *file, char __user *buf, size_t len, loff_t *ppos)
{
	ssize_t r;
	inode_lock_shared(file->f_inode);
	r = nvfs_read_locked(file, buf, len, ppos);
	inode_unlock_shared(file->f_inode);
	return r;
}

static ssize_t nvfs_write(struct file *file, const char __user *buf, size_t len, loff_t *ppos)
{
	ssize_t r;
	inode_lock(file->f_inode);
	r = nvfs_write_locked(file, buf, len, ppos);
	inode_unlock(file->f_inode);
	return r;
}

static ssize_t nvfs_rw_iter(struct kiocb *iocb, struct iov_iter *iov)
{
	ssize_t r = 0;
	ssize_t total = 0;
	bool wr = iov->type & WRITE;

	if (!wr)
		inode_lock_shared(iocb->ki_filp->f_inode);
	else
		inode_lock(iocb->ki_filp->f_inode);

	while (iov_iter_count(iov)) {
		struct iovec iovec = iov_iter_iovec(iov);

		if (!wr) {
			r = nvfs_read_locked(iocb->ki_filp, iovec.iov_base, iovec.iov_len, &iocb->ki_pos);
		} else {
			r = nvfs_write_locked(iocb->ki_filp, iovec.iov_base, iovec.iov_len, &iocb->ki_pos);
		}
		if (unlikely(r <= 0))
			break;
		total += r;
		if (unlikely(r != iovec.iov_len))
			break;

		iov_iter_advance(iov, r);
	}

	if (!wr)
		inode_unlock_shared(iocb->ki_filp->f_inode);
	else
		inode_unlock(iocb->ki_filp->f_inode);

	return total ? total : r;
}

int nvfs_fsync(struct file *file, loff_t start, loff_t end, int datasync)
{
	return 0;
}

static long nvfs_fallocate(struct file *file, int mode, loff_t offset, loff_t len)
{
	struct inode *inode = file_inode(file);
	int r;

	if (mode & FALLOC_FL_PUNCH_HOLE) {
		inode_lock(inode);
		r = nvfs_punch_hole(i_to_nmi(inode), offset, offset + len);
		inode_unlock(inode);

		return r;
	}

	return -EOPNOTSUPP;
}

const struct inode_operations nvfs_file_inode_operations = {
	.setattr = nvfs_setattr,
	.update_time = nvfs_update_time,
};

const struct file_operations nvfs_file_fops = {
	.llseek = generic_file_llseek,
	.read = nvfs_read,
	.write = nvfs_write,
	.read_iter = nvfs_rw_iter,
	.write_iter = nvfs_rw_iter,
	.fsync = nvfs_fsync,
	.fallocate = nvfs_fallocate,
	.mmap = nvfs_mmap,
};
