#include "nvfs.h"

static const struct super_operations nvfs_sops;
static struct kmem_cache *nvfs_inode_cache;

static int persistent_memory_claim(struct nvfs_superblock *nvs)
{
	int r;
	loff_t s;
	long p, da;
	pfn_t pfn;
	int id;
	struct page **pages;

	nvs->memory_map_size = i_size_read(nvs->s->s_bdev->bd_inode);
	s = nvs->memory_map_size;
	p = s >> PAGE_SHIFT;
	if (!p) {
		r = -EINVAL;
		goto err1;
	}
	if (p != s >> PAGE_SHIFT) {
		r = -EOVERFLOW;
		goto err1;
	}

	nvs->dax_dev = fs_dax_get_by_bdev(nvs->s->s_bdev);
	if (!nvs->dax_dev) {
		r = -EOPNOTSUPP;
		goto err1;
	}

	id = dax_read_lock();

	da = dax_direct_access(nvs->dax_dev, 0, p, &nvs->memory_map, &pfn);
	if (da < 0) {
		nvs->memory_map = NULL;
		r = da;
		goto err2;
	}
	if (!pfn_t_has_page(pfn)) {
		nvs->memory_map = NULL;
		r = -EOPNOTSUPP;
		goto err2;
	}
	if (da != p) {
		long i;
		nvs->memory_map = NULL;
		pages = kvmalloc_array(p, sizeof(struct page *), GFP_KERNEL);
		if (!pages) {
			r = -ENOMEM;
			goto err2;
		}
		i = 0;
		do {
			long daa;
			daa = dax_direct_access(nvs->dax_dev, i, p - i, NULL, &pfn);
			if (daa <= 0) {
				r = daa ? daa : -EINVAL;
				goto err3;
			}
			if (!pfn_t_has_page(pfn)) {
				r = -EOPNOTSUPP;
				goto err3;
			}
			while (daa-- && i < p) {
				pages[i++] = pfn_t_to_page(pfn);
				pfn.val++;
			}
		} while (i < p);
		nvs->memory_map = vmap(pages, p, VM_MAP, PAGE_KERNEL);
		if (!nvs->memory_map) {
			r = -ENOMEM;
			goto err3;
		}
		kvfree(pages);
		nvs->memory_vmapped = true;
	}

	dax_read_unlock(id);

	return 0;
err3:
	kvfree(pages);
err2:
	dax_read_unlock(id);
err1:
	return r;
}

__cold static void persistent_memory_release(struct nvfs_superblock *nvs)
{
	if (nvs->memory_vmapped)
		vunmap(nvs->memory_map);
	if (nvs->dax_dev)
		put_dax(nvs->dax_dev);
}

void nvfs_error(struct nvfs_superblock *nvs, const char *fmt, ...)
{
	static DEFINE_RATELIMIT_STATE(nvfs_error_ratelimit, 5 * HZ, 10);
	struct va_format vaf;
	va_list args;

	if (!__ratelimit(&nvfs_error_ratelimit))
		goto skip_print;

	va_start(args, fmt);
	vaf.fmt = fmt;
	vaf.va = &args;

	printk(KERN_CRIT "NVFS (%s): error: %pV\n", nvs->s->s_id, &vaf);
	dump_stack();

	va_end(args);

skip_print:
	return;
}

static int nvfs_statfs(struct dentry *dentry, struct kstatfs *buf)
{
	struct nvfs_superblock *nvs = dentry->d_sb->s_fs_info;

	uint64_t fsid;

	buf->f_type = le32_to_cpu(*(uint32_t *)NVFS_MAGIC);
	buf->f_bsize = 1UL << nvs->log2_page_size;
	buf->f_blocks = le64_to_cpu(p_superblock(nvs)->n_pages);
	buf->f_bfree = percpu_counter_sum_positive(&nvs->free_blocks_cnt);
	buf->f_bavail = buf->f_bfree;
	buf->f_files = nvs->n_groups << nvs->log2_inodes_per_group;
	buf->f_ffree = percpu_counter_sum_positive(&nvs->free_inodes_cnt);
	buf->f_namelen = (1UL << nvs->log2_page_size) - offsetof(struct nvfs_dir_entry, name);

	fsid =	le64_to_cpu(*(uint64_t *)(p_superblock(nvs)->uuid)) ^
		le64_to_cpu(*(uint64_t *)(p_superblock(nvs)->uuid + 8));
	buf->f_fsid.val[0] = fsid & 0xFFFFFFFFUL;
	buf->f_fsid.val[1] = (fsid >> 32) & 0xFFFFFFFFUL;

	return 0;
}

static void nvfs_inode_ctor(void *n)
{
	struct nvfs_memory_inode *nmi = n;
	spin_lock_init(&nmi->metadata_lock);
	init_rwsem(&nmi->dax_sem);
	inode_init_once(&nmi->vfs_inode);
}

static struct inode *nvfs_cache_alloc_inode(struct super_block *sb)
{
	struct nvfs_memory_inode *nmi;
	nmi = kmem_cache_alloc(nvfs_inode_cache, GFP_NOIO);
	if (unlikely(!nmi))
		return NULL;
	return &nmi->vfs_inode;
}

static void nvfs_cache_free_inode(struct inode *inode)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	kmem_cache_free(nvfs_inode_cache, nmi);
}

__cold static void nvfs_free_super(struct nvfs_superblock *nvs)
{
	if (!nvs)
		return;
	persistent_memory_release(nvs);
	percpu_counter_destroy(&nvs->free_blocks_cnt);
	percpu_counter_destroy(&nvs->free_inodes_cnt);
	free_percpu(nvs->npc);
	kvfree(nvs->group_info);
	kfree(nvs);
}

__cold static void nvfs_put_super(struct super_block *s)
{
	struct nvfs_superblock *nvs = s->s_fs_info;
	uint64_t group;
	for (group = 0; group < nvs->n_groups; group++) {
		struct nvfs_group *p = p_group(nvs, group);
		arch_wb_cache_pmem(p, sizeof(struct nvfs_group));
		cond_resched();
	}
	nvfs_free_super(nvs);
}

__cold static int nvfs_fill_super(struct super_block *s, void *options, int silent)
{
	struct nvfs_superblock *nvs;
	struct inode *root;
	int r;
	unsigned c;
	unsigned long i;
	const char *msg;
	uint64_t max_ino, max_block;
	uint64_t free_blocks, free_inodes;

	nvs = kzalloc(sizeof(struct nvfs_superblock), GFP_KERNEL);
	if (!nvs) {
		r = -ENOMEM;
		goto err0;
	}

	nvs->s = s;
	s->s_fs_info = nvs;
	s->s_op = &nvfs_sops;
	s->s_maxbytes = MAX_LFS_FILESIZE;
	s->s_time_min = S64_MIN;
	s->s_time_max = S64_MAX;
	s->s_time_gran = 1;

	r = persistent_memory_claim(nvs);
	if (r)
		goto err0;

	msg = nvfs_validate_super(nvs->memory_map, nvs->memory_map_size);
	if (msg) {
		printk(KERN_ERR "nvfs: %s", msg);
		r = -EINVAL;
		goto err0;
	}

	nvs->log2_page_size = p_superblock(nvs)->log2_page_size;
	nvs->log2_inode_size = p_superblock(nvs)->log2_inode_size;
	nvs->log2_bytes_per_group = p_superblock(nvs)->log2_page_size * 2 + 3;
	nvs->log2_inodes_per_group = p_superblock(nvs)->log2_inode_pages_per_group + p_superblock(nvs)->log2_page_size - p_superblock(nvs)->log2_inode_size;

	nvfs_get_sizes(p_superblock(nvs), &nvs->n_groups, NULL, NULL, NULL, NULL);

	nvs->npc = alloc_percpu(struct nvfs_per_cpu);
	if (!nvs->npc) {
		r = -ENOMEM;
		goto err0;
	}
	max_ino = nvs->n_groups << nvs->log2_inodes_per_group;
	max_block = nvs->n_groups << log2_pages_per_group(nvs);
	for_each_possible_cpu(c) {
		struct nvfs_per_cpu *npc = per_cpu_ptr(nvs->npc, c);
		npc->current_inode = nvfs_get_random(max_ino);
		npc->current_block = nvfs_get_random(max_block);
	}

	nvfs_count_objects(nvs, &free_blocks, &free_inodes);
	r = percpu_counter_init(&nvs->free_blocks_cnt, free_blocks, GFP_KERNEL);
	if (r) {
		r = -ENOMEM;
		goto err0;
	}
	r = percpu_counter_init(&nvs->free_inodes_cnt, free_inodes, GFP_KERNEL);
	if (r) {
		r = -ENOMEM;
		goto err0;
	}

	nvs->group_info = kvmalloc_array(nvs->n_groups, sizeof(struct nvfs_group_info), GFP_KERNEL);
	if (!nvs->group_info) {
		r = -ENOMEM;
		goto err0;
	}

	for (i = 0; i < nvs->n_groups; i++) {
		struct nvfs_group_info *gi = nvs->group_info + i;
		struct nvfs_group *g = p_group(nvs, i);

		/* !!! TODO: recalculate this if we have crashed */

		gi->n_free_blocks = le32_to_cpu(g->n_free_blocks);
		gi->n_free_inodes = le32_to_cpu(g->n_free_inodes);

		spin_lock_init(&gi->lock);
	}

	root = nvfs_iget(s, le64_to_cpu(p_superblock(nvs)->root_inode));
	if (IS_ERR(root)) {
		r = PTR_ERR(root);
		goto err0;
	}
	if (!S_ISDIR(root->i_mode)) {
		iput(root);
		r = -EUCLEAN;
		goto err0;
	}
	s->s_root = d_make_root(root);
	if (!s->s_root) {
		r = -ENOMEM;
		goto err0;
	}

	return 0;

err0:
	nvfs_free_super(nvs);
	return r;
}

__cold static struct dentry *nvfs_mount(struct file_system_type *fs_type, int flags,
					const char *dev_name, void *data)
{
	return mount_bdev(fs_type, flags, dev_name, data, nvfs_fill_super);
}

static const struct super_operations nvfs_sops = {
	.alloc_inode = nvfs_cache_alloc_inode,
	.free_inode = nvfs_cache_free_inode,
	.evict_inode = nvfs_evict_inode,
	.put_super = nvfs_put_super,
	.statfs  = nvfs_statfs,
};

static struct file_system_type nvfs_fs_type = {
	.owner = THIS_MODULE,
	.name = "nvfs",
	.mount = nvfs_mount,
	.kill_sb = kill_block_super,
	.fs_flags = FS_REQUIRES_DEV
};

__cold static int __init nvfs_init(void)
{
	int r;

	nvfs_inode_cache = kmem_cache_create("nvfs_inode_cache", sizeof(struct nvfs_memory_inode), 0, SLAB_RECLAIM_ACCOUNT | SLAB_MEM_SPREAD | SLAB_ACCOUNT, nvfs_inode_ctor);
	if (!nvfs_inode_cache) {
		r = -ENOMEM;
		goto ret_r;
	}

	r = register_filesystem(&nvfs_fs_type);
	if (unlikely(r))
		goto ret_r;

	return 0;

ret_r:
	kmem_cache_destroy(nvfs_inode_cache);
	return r;
}

__cold static void __exit nvfs_exit(void)
{
	unregister_filesystem(&nvfs_fs_type);
	rcu_barrier();
	kmem_cache_destroy(nvfs_inode_cache);
}

module_init(nvfs_init)
module_exit(nvfs_exit)

MODULE_ALIAS_FS("nvfs");

MODULE_AUTHOR("Mikulas Patocka");
MODULE_DESCRIPTION("NVFS Filesystem");
MODULE_LICENSE("GPL");
